package com.example.ponguoc;

import java.io.IOException;

import org.andengine.engine.camera.Camera;
import org.andengine.engine.handler.IUpdateHandler;
import org.andengine.engine.options.EngineOptions;
import org.andengine.engine.options.ScreenOrientation;
import org.andengine.engine.options.resolutionpolicy.RatioResolutionPolicy;
import org.andengine.entity.primitive.Line;
import org.andengine.entity.primitive.Rectangle;
import org.andengine.entity.scene.IOnSceneTouchListener;
import org.andengine.entity.scene.Scene;
import org.andengine.entity.scene.menu.MenuScene;
import org.andengine.entity.text.Text;
import org.andengine.entity.util.FPSLogger;
import org.andengine.extension.multiplayer.adt.message.IMessage;
import org.andengine.extension.multiplayer.util.MessagePool;
import org.andengine.input.touch.TouchEvent;
import org.andengine.opengl.font.Font;
import org.andengine.opengl.font.FontFactory;
import org.andengine.opengl.vbo.VertexBufferObjectManager;
import org.andengine.ui.activity.SimpleBaseGameActivity;
import org.andengine.util.WifiUtils;
import org.andengine.util.WifiUtils.WifiUtilsException;

import android.app.AlertDialog;
import android.app.Dialog;
import android.content.DialogInterface;
import android.content.DialogInterface.OnClickListener;
import android.graphics.Color;
import android.os.Bundle;
import android.util.SparseArray;
import android.view.KeyEvent;
import android.widget.EditText;
import android.widget.Toast;

import com.example.ponguoc.GameClientMessages.ClientMsgFinishGame;
import com.example.ponguoc.GameClientMessages.ClientMsgMovePaddle;
import com.example.ponguoc.GameClientMessages.ClientMsgPauseGame;
import com.example.ponguoc.GameClientMessages.ClientMsgRestartGame;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateBall;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdatePaddle;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateScore;

public class GameActivity extends SimpleBaseGameActivity implements IOnSceneTouchListener{
	
	public static final int CAMERA_WIDTH = 480;
	public static final int CAMERA_HEIGHT = 800;
	public static final int IS_SERVER = 0;
	public static final int IS_CLIENT = 1;
	public static final int IS_SINGLE = 2;
	public static final int FPS = 30;
	private static final String LOCALHOST_IP = "127.0.0.1";
	public static final int SERVER_PORT = 4443;
	public static final int PADDLE_WIDTH = 80;
	public static final int PADDLE_HEIGHT = 20;
	public static final int BALL_SIDE = 15;
	public static final int TABLE_WIDTH = MainActivity.CAMERA_WIDTH;
	public static final int TABLE_HEIGHT = MainActivity.CAMERA_HEIGHT;
	
	public Camera mCamera;
	public Options mOptions;
	public GameClient mClient;
	public GameServer mServer;
	private String mServerIP = LOCALHOST_IP;
	
	private final MessagePool<IMessage> mMessagePool = new MessagePool<IMessage>();
	public Font mFontScore;
	public Font mFontMenu;
	
	public Options gameOptions;
	public Rectangle mBall;
	//public Paddle myPaddle, opponentPaddle;
	public SparseArray<Paddle> mPaddleList = new SparseArray<Paddle>();
	//public Score myScore, opponentScore;
	public SparseArray<Text> mScoreList = new SparseArray<Text>();
	public MainActivity activity;
	public float mMyPaddleYPos;
	public int mPaddleID;
	public int mType;
	public Scene mScene;
	public MenuScene mPauseScene;

	
	public static GameActivity instance;

	@Override
	public EngineOptions onCreateEngineOptions() {

		instance = this;
		
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_BALL, ServerMsgUpdateBall.class);
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_PADDLE, ServerMsgUpdatePaddle.class);
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_SCORE, ServerMsgUpdateScore.class);
		this.mMessagePool.registerMessage(GameClientMessages.CLIENT_MESSAGE_MOVE_PADDLE, ClientMsgMovePaddle.class);
		this.mMessagePool.registerMessage(GameClientMessages.CLIENT_MESSAGE_PAUSE_GAME, ClientMsgPauseGame.class);
		this.mMessagePool.registerMessage(GameClientMessages.CLIENT_MESSAGE_RESTART_GAME, ClientMsgRestartGame.class);
		this.mMessagePool.registerMessage(GameClientMessages.CLIENT_MESSAGE_FINISH_GAME, ClientMsgFinishGame.class);
		
		mCamera = new Camera(0, 0, CAMERA_WIDTH, CAMERA_HEIGHT);
		
	    return new EngineOptions(true, ScreenOrientation.PORTRAIT_SENSOR,
	    		new RatioResolutionPolicy(CAMERA_WIDTH, CAMERA_HEIGHT), mCamera);
	}

	@Override
	protected void onCreateResources() throws IOException {
	    mFontScore = FontFactory.createFromAsset(this.getFontManager(),this.getTextureManager(),
				256, 256, this.getAssets(), "fonts/LCD.ttf", 36f, true, Color.RED);
	    mFontScore.load();
	    
	    mFontMenu = FontFactory.createFromAsset(this.getFontManager(),this.getTextureManager(),
				256, 256, this.getAssets(), "fonts/stocky.ttf", 36f, true, Color.RED);
	    mFontMenu.load();
		
	}

	@Override
	protected Scene onCreateScene() {
		this.mEngine.registerUpdateHandler(new FPSLogger());
		final VertexBufferObjectManager vertexBufferObjectManager = this.getVertexBufferObjectManager();
		Bundle b = new Bundle();
		b = getIntent().getExtras();
		mType = b.getInt("type");
		mOptions = b.getParcelable("options");
		mScene = new Scene();
		//create ball
		mBall = new Rectangle(GameActivity.CAMERA_WIDTH/2,
				GameActivity.CAMERA_HEIGHT/2, BALL_SIDE, BALL_SIDE, vertexBufferObjectManager);
		mBall.setPosition(TABLE_WIDTH/2, TABLE_HEIGHT/2);
		mScene.attachChild(mBall);

		//create table lines
		mScene.attachChild(new Line(0, 0, 0, TABLE_HEIGHT, vertexBufferObjectManager)); //left line
		mScene.attachChild(new Line(0, TABLE_HEIGHT, TABLE_WIDTH, TABLE_HEIGHT, vertexBufferObjectManager)); //top line
		mScene.attachChild(new Line(TABLE_WIDTH-1, TABLE_HEIGHT, TABLE_WIDTH-1, 0, vertexBufferObjectManager)); //right line
		mScene.attachChild(new Line(TABLE_WIDTH, 1, 0, 1, vertexBufferObjectManager)); //bottom line
		mScene.attachChild(new Line(0, TABLE_HEIGHT/2, TABLE_WIDTH, TABLE_HEIGHT/2, vertexBufferObjectManager)); //centre line

		//create paddles
		Paddle paddleBottom = new Paddle(PADDLE_WIDTH, PADDLE_HEIGHT);
		Paddle paddleTop = new Paddle(PADDLE_WIDTH, PADDLE_HEIGHT);
		// if single player, paddle covering all goal line
		if(mType == IS_SINGLE) paddleTop = new Paddle(CAMERA_WIDTH, PADDLE_HEIGHT);
		paddleBottom.setPosition(TABLE_WIDTH/2, 20);
		paddleTop.setPosition(TABLE_WIDTH/2, TABLE_HEIGHT-20);
		mPaddleList.put(IS_SERVER, paddleBottom);
		mPaddleList.put(IS_CLIENT, paddleTop);

		mScene.attachChild(paddleBottom);
		mScene.attachChild(paddleTop);

		//create score
		Text scoreBottom = new Text(0, 0, mFontScore, "0 ",vertexBufferObjectManager);
		Text scoreTop = new Text(0, 0, mFontScore, "0 ",vertexBufferObjectManager);
		scoreBottom.setPosition(TABLE_WIDTH - 30, TABLE_HEIGHT/2 - 30);
		scoreTop.setPosition(30, TABLE_HEIGHT/2 + 30);
		mScoreList.put(IS_SERVER, scoreBottom);
		mScoreList.put(IS_CLIENT, scoreTop);

		mScene.attachChild(scoreBottom);
		mScene.attachChild(scoreTop);
		
		mPaddleID = mType;
		if(mPaddleID == IS_SINGLE){ //paddleID has to be 0 if player is single
			mPaddleID = IS_SERVER;
		}
		runOnUiThread(new Runnable(){
			public void run(){
				showDialog(mType);
			}
		});

		mScene.setOnSceneTouchListener(this);

		mScene.registerUpdateHandler(new IUpdateHandler() {
			@Override
			public void onUpdate(final float pSecondsElapsed) {
				if(mClient != null){
					ClientMsgMovePaddle message = (ClientMsgMovePaddle) 
							mMessagePool.obtainMessage(GameClientMessages.CLIENT_MESSAGE_MOVE_PADDLE);
					message.setPaddleID(mPaddleID, mMyPaddleYPos);
					mClient.sendMessage(message);
					mMessagePool.recycleMessage(message);
					//Log.i("pong", "onUpdate() entered, message sent");
				}
				//else Log.i("pong", "onUpdate() entered, client still not created");
			}

			@Override
			public void reset() {}
		});
		return mScene;
	}

	@Override
	public boolean onSceneTouchEvent(Scene pScene, TouchEvent pSceneTouchEvent) {
		//Log.i("pong", "Touch event in GameScene (paddle listener)");
		this.mMyPaddleYPos = pSceneTouchEvent.getX();
		if(mServer != null){
			mServer.setStartGame(true);
		}
		return true;
	}
	
	public void updatePaddle(final int paddleID, final float pX, final float pY) {
		this.mPaddleList.get(paddleID).setPosition(pX, pY);
	}

	public void updateBall(final float pX, final float pY) {
		this.mBall.setPosition(pX, pY);
	}
	
	public void updateScore(final int paddleID, final int points){
		this.mScoreList.get(paddleID).setText(String.valueOf(points));
	}
	
	public int getType(){
		return this.mType;
	}
	
	public int getPaddleID(){
		return this.mPaddleID;
	}
	
	public float getMyPaddleYPos(){
		return this.mMyPaddleYPos;
	}
	
	@Override
	protected Dialog onCreateDialog(int id) {
		switch (id) {
		// Create dialog that allows user to choose either server, client,
		// or both
		case IS_SERVER:
			this.initServerAndClient(mType);
			
			try{
			return new AlertDialog.Builder(this)
					.setTitle("Your server IP:")
					.setMessage("The IP of your Server is:\n" + WifiUtils.getWifiIPv4Address(this))
					.setCancelable(false)
					.setPositiveButton("OK", new OnClickListener() {
						@Override
						public void onClick(final DialogInterface pDialog, final int pWhich) {
							toast("Touch the screen to start the game");
						}
					})
					.create();
			}catch (final WifiUtilsException e) {
				return new AlertDialog.Builder(this)
				.setIcon(android.R.drawable.ic_dialog_alert)
				.setTitle("Your Server-IP ...")
				.setCancelable(false)
				.setMessage("Error retrieving IP of your Server: " + e)
				.setPositiveButton(android.R.string.ok, new OnClickListener() {
					@Override
					public void onClick(final DialogInterface pDialog, final int pWhich) {
						finish();
					}
				})
				.create();
			}
		// Create the 'Enter IP' dialog, allowing users to enter a server
		// IP to connect to if they select to be solely a client
		case IS_CLIENT:
			final EditText editText = new EditText(this);
			return new AlertDialog.Builder(this).setTitle("Enter Server IP...")
					.setCancelable(false).setView(editText)
					.setPositiveButton("Connect", new OnClickListener() {

						@Override
						public void onClick(DialogInterface dialog, int which) {
							
							// Set the mServerIP String to that of the edit text box
							mServerIP = editText.getText().toString();
							
							// Create a new client object, passing the server, port,
							// engine and scene to our multiplayer client
							mClient = new GameClient(mServerIP,
									SERVER_PORT, mEngine, mScene);
							
							// Initialize the client
							mClient.initClient();
						}
					}).create();
		case IS_SINGLE:
			this.initServerAndClient(mType);
			return new AlertDialog.Builder(this)
					.setTitle("Single Mode")
					.setMessage("How much time can you resist without receiving a goal?")
					.setCancelable(false)
					.setPositiveButton("Start!", new OnClickListener() {
						@Override
						public void onClick(final DialogInterface pDialog, final int pWhich) {
							toast("Touch the screen to start the game");
						}
					})
					.create();
		}
		
		return super.onCreateDialog(id);
	}
	
	private void initServerAndClient(int pType) {
		
		// Create & initialize the server
		boolean isSingle = false;
		if(pType == IS_SINGLE) isSingle = true;
		mServer = new GameServer(SERVER_PORT, mEngine, mOptions, isSingle);
		mServer.initServer();

		// We're allowing the thread to sleep for
		// a little bit to give the server time to start up
		// before initializing the client
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}

		// Create & initialize the server
		mClient = new GameClient(LOCALHOST_IP, SERVER_PORT, mEngine, mScene);
		mClient.initClient();
		
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}

	@Override
	protected void onDestroy() {

		if(mClient != null) {
			mClient.terminate();
		}
		if(mServer != null){
			mServer.terminate();
		}
		
		mEngine.stop();

		super.onDestroy();
	}
	public static GameActivity getSharedInstance(){
		return instance;
	}
	
	public void toast(final String pMessage) {
		this.runOnUiThread(new Runnable() {
			@Override
			public void run() {
				Toast.makeText(GameActivity.this, pMessage, Toast.LENGTH_SHORT).show();
			}
		});
	}
	
	@Override
	public void onBackPressed(){
		if(mClient != null){
			mClient.terminate();
		}
		if(mServer != null){
			mServer.terminate();
		}
		finish();
	}
	
	public void finishGameActivity(){
		if(mClient != null){
			mClient.terminate();
		}
		if(mServer != null){
			mServer.terminate();
		}
		finish();
	}
	
	public void resumePausedGame(){
		//mPauseScene.closeMenuScene();
		if(mClient != null){
			ClientMsgPauseGame message = (ClientMsgPauseGame) 
					mMessagePool.obtainMessage(GameClientMessages.CLIENT_MESSAGE_PAUSE_GAME);
			message.setPaused(false);
			mClient.sendMessage(message);
			mMessagePool.recycleMessage(message);
		}
	}
	
	@Override
	public boolean onKeyDown(int keyCode, KeyEvent event){
		if (keyCode == KeyEvent.KEYCODE_MENU) {
			if(mClient != null){
				ClientMsgPauseGame message = (ClientMsgPauseGame) 
						mMessagePool.obtainMessage(GameClientMessages.CLIENT_MESSAGE_PAUSE_GAME);
				message.setPaused(true);
				mClient.sendMessage(message);
				mMessagePool.recycleMessage(message);
				//Log.i("pong", "onUpdate() entered, message sent");
			}
			mPauseScene = new PauseScene();
			mScene.setChildScene(mPauseScene);
            return true;
        }
		return super.onKeyDown(keyCode, event); 
	}
	
	public void sendFinishMessage(){
		if(mClient != null){
			ClientMsgFinishGame message = (ClientMsgFinishGame) 
					mMessagePool.obtainMessage(GameClientMessages.CLIENT_MESSAGE_FINISH_GAME);
			mClient.sendMessage(message);
			mMessagePool.recycleMessage(message);
			//Log.i("pong", "onUpdate() entered, message sent");
		}
	}
	
	public void sendRestartMessage(){
		if(mClient != null){
			ClientMsgRestartGame message = (ClientMsgRestartGame) 
					mMessagePool.obtainMessage(GameClientMessages.CLIENT_MESSAGE_RESTART_GAME);
			mClient.sendMessage(message);
			mMessagePool.recycleMessage(message);
			//Log.i("pong", "onUpdate() entered, message sent");
		}
	}
	

}
