package com.example.ponguoc;

import java.io.IOException;
import java.net.Socket;

import org.andengine.engine.Engine;
import org.andengine.entity.scene.Scene;
import org.andengine.extension.multiplayer.adt.message.client.ClientMessage;
import org.andengine.extension.multiplayer.adt.message.server.IServerMessage;
import org.andengine.extension.multiplayer.client.IServerMessageHandler;
import org.andengine.extension.multiplayer.client.connector.ServerConnector;
import org.andengine.extension.multiplayer.client.connector.SocketConnectionServerConnector;
import org.andengine.extension.multiplayer.client.connector.SocketConnectionServerConnector.ISocketConnectionServerConnectorListener;
import org.andengine.extension.multiplayer.shared.SocketConnection;

import android.util.Log;

import com.example.ponguoc.GameServerMessages.ServerMsgFinishGame;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateBall;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdatePaddle;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateScore;

public class GameClient implements ISocketConnectionServerConnectorListener{

	private Engine mEngine;
	private Scene mScene;
	private GameActivity activity = GameActivity.getSharedInstance();
	
	// Server variables
	private String mServerIP;
	private int mServerPort;
	
	// ServerConnector object which deals with the communication between client --> server
	private ServerConnector<SocketConnection> mServerConnector;
	
	public GameClient(final String pServerIP, final int pServerPort, final Engine pEngine, final Scene pScene){
		this.mServerIP = pServerIP;
		this.mServerPort = pServerPort;
		this.mEngine = pEngine;
		this.mScene = pScene;
	}
	
	public void initClient(){
		
		Log.i("pong", "Enter initClient()");
		// A separate thread is needed for network communication
		// in order to avoid blocking on the main thread
		this.mEngine.runOnUpdateThread(new Runnable(){

			@Override
			public void run() {
				try {
					
					// Create the socket with the specified Server IP and port
					Socket socket = new Socket(GameClient.this.mServerIP, GameClient.this.mServerPort);

					// Create the socket connection, establishing the input/output stream
					SocketConnection socketConnection = new SocketConnection(socket);
					// Create the server connector with the specified socket connection
					// and client connection listener
					mServerConnector = new SocketConnectionServerConnector(socketConnection, GameClient.this);
					Log.i("pong", "client mServerConnector created");
					// Register server messages to the client with a message handler
					mServerConnector.registerServerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_BALL,
							ServerMsgUpdateBall.class, new IServerMessageHandler<SocketConnection>(){
						
						// If a client receives the SERVER_MESSAGE_UPDATE_BALL server message,
						// Fire the following code...
						@Override
						public void onHandleMessage(
								ServerConnector<SocketConnection> pServerConnector,
								IServerMessage pServerMessage)
								throws IOException {
							
							// obtain the class casted server message
							ServerMsgUpdateBall message = (ServerMsgUpdateBall) pServerMessage;
							
							// Update ball, based on values obtained via the server
							// message received
							GameActivity.getSharedInstance().updateBall(message.getX(), message.getY());
						}
						
					});
					Log.i("pong", "client updateball msg registered");
					
					mServerConnector.registerServerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_PADDLE,
							ServerMsgUpdatePaddle.class, new IServerMessageHandler<SocketConnection>(){

								@Override
								public void onHandleMessage(
										ServerConnector<SocketConnection> pServerConnector,
										IServerMessage pServerMessage)
										throws IOException {
									
									// obtain the class casted server message
									ServerMsgUpdatePaddle message = (ServerMsgUpdatePaddle) pServerMessage;
									
									// Update ball, based on values obtained via the server
									// message received
									GameActivity.getSharedInstance().updatePaddle(message.mPaddleID, message.mX, message.mY);
								}
					});
					Log.i("pong", "client updatepaddle msg registered");
					
					mServerConnector.registerServerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_SCORE,
							ServerMsgUpdateScore.class, new IServerMessageHandler<SocketConnection>(){

								@Override
								public void onHandleMessage(
										ServerConnector<SocketConnection> pServerConnector,
										IServerMessage pServerMessage)
										throws IOException {
									
									// obtain the class casted server message
									ServerMsgUpdateScore message = (ServerMsgUpdateScore) pServerMessage;
									
									// Update ball, based on values obtained via the server
									// message received
									GameActivity.getSharedInstance().updateScore(message.mPaddleID, message.mScore);
								}
					});
					Log.i("pong", "client updatescore msg registered");
					
					mServerConnector.registerServerMessage(GameServerMessages.SERVER_MESSAGE_FINISH_GAME,
							ServerMsgFinishGame.class, new IServerMessageHandler<SocketConnection>(){

								@Override
								public void onHandleMessage(
										ServerConnector<SocketConnection> pServerConnector,
										IServerMessage pServerMessage)
										throws IOException {
									
									activity.finishGameActivity();
								}
					});
					
					// Once we've created our server connector and registered our server messages,
					// we can call start() on the server connector's connection
					mServerConnector.getConnection().start();
					Log.i("pong", "Client initialized and started");
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}, true);
	}
	
	public void sendMessage(ClientMessage pClientMessage){
		try {
			this.mServerConnector.sendClientMessage(pClientMessage);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public void terminate(){
		if(this.mServerConnector != null){
			Log.i("pong", "CLIENT: backbutton pressed, serverconnector terminated");
			this.mServerConnector.terminate();
		}
	}
	
	@Override
	public void onStarted(ServerConnector<SocketConnection> pServerConnector) {
		activity.toast("CLIENT: connected to server");
		Log.i("pong", "CLIENT: connected to server");
		
	}

	@Override
	public void onTerminated(ServerConnector<SocketConnection> pServerConnector) {
		activity.toast("CLIENT: disconnected from server");
		//mServerConnector.terminate();
		
	}

}
