package com.example.ponguoc;

import java.io.IOException;
import java.util.ArrayList;

import org.andengine.engine.Engine;
import org.andengine.engine.handler.IUpdateHandler;
import org.andengine.extension.multiplayer.adt.message.IMessage;
import org.andengine.extension.multiplayer.adt.message.client.IClientMessage;
import org.andengine.extension.multiplayer.server.IClientMessageHandler;
import org.andengine.extension.multiplayer.server.SocketServer;
import org.andengine.extension.multiplayer.server.SocketServer.ISocketServerListener;
import org.andengine.extension.multiplayer.server.connector.ClientConnector;
import org.andengine.extension.multiplayer.server.connector.SocketConnectionClientConnector;
import org.andengine.extension.multiplayer.server.connector.SocketConnectionClientConnector.ISocketConnectionClientConnectorListener;
import org.andengine.extension.multiplayer.shared.SocketConnection;
import org.andengine.extension.multiplayer.util.MessagePool;
import org.andengine.extension.physics.box2d.FixedStepPhysicsWorld;
import org.andengine.extension.physics.box2d.PhysicsFactory;
import org.andengine.extension.physics.box2d.PhysicsWorld;
import org.andengine.extension.physics.box2d.util.Vector2Pool;
import org.andengine.extension.physics.box2d.util.constants.PhysicsConstants;
import org.andengine.util.adt.list.SmartList;
import org.andengine.util.debug.Debug;
import org.andengine.util.math.MathUtils;

import android.util.Log;
import android.util.SparseArray;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.physics.box2d.Body;
import com.badlogic.gdx.physics.box2d.BodyDef.BodyType;
import com.badlogic.gdx.physics.box2d.Contact;
import com.badlogic.gdx.physics.box2d.ContactImpulse;
import com.badlogic.gdx.physics.box2d.ContactListener;
import com.badlogic.gdx.physics.box2d.Fixture;
import com.badlogic.gdx.physics.box2d.FixtureDef;
import com.badlogic.gdx.physics.box2d.Manifold;
import com.example.ponguoc.GameClientMessages.ClientMsgFinishGame;
import com.example.ponguoc.GameClientMessages.ClientMsgMovePaddle;
import com.example.ponguoc.GameClientMessages.ClientMsgPauseGame;
import com.example.ponguoc.GameClientMessages.ClientMsgRestartGame;
import com.example.ponguoc.GameServerMessages.ServerMsgFinishGame;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateBall;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdatePaddle;
import com.example.ponguoc.GameServerMessages.ServerMsgUpdateScore;

public class GameServer implements IUpdateHandler, ContactListener,
ISocketServerListener<SocketConnectionClientConnector>, ISocketConnectionClientConnectorListener{
	
	private static final FixtureDef PADDLE_FIXTUREDEF = PhysicsFactory.createFixtureDef(1, 1, 0);
	private static final FixtureDef BALL_FIXTUREDEF = PhysicsFactory.createFixtureDef(1, 1, 0);
	private static final FixtureDef WALL_FIXTUREDEF = PhysicsFactory.createFixtureDef(1, 1, 0);
	
	private final PhysicsWorld mPhysicsWorld;
	private final Body mBallBody;
	private final SparseArray<Body> mPaddleBodies = new SparseArray<Body>();
	private boolean mResetBall = true;
	private final SparseArray<Score> mPaddleScores = new SparseArray<Score>();

	private final MessagePool<IMessage> mMessagePool = new MessagePool<IMessage>();
	private final ArrayList<ServerMsgUpdatePaddle> mUpdatePaddleServerMessages = new ArrayList<ServerMsgUpdatePaddle>();
	
	private int mServerPort;
	private Engine mEngine;
	public SocketServer<SocketConnectionClientConnector> mSocketServer;
	private Body mTopGoal;
	private boolean mStartGame = false;
	public Options mOptions;
	public GameActivity activity = GameActivity.getSharedInstance();

	public GameServer(int pServerPort, Engine pEngine) {

		Log.i("pong", "Server constructor entered");
		this.mServerPort = pServerPort;
		this.mEngine = pEngine;
		
		this.mEngine.registerUpdateHandler(this);
		this.initMessagePool();

		this.mPaddleScores.put(MainActivity.IS_SERVER, new Score());
		this.mPaddleScores.put(MainActivity.IS_CLIENT, new Score());

		this.mPhysicsWorld = new FixedStepPhysicsWorld(MainActivity.FPS, 2, new Vector2(0, 0), false, 8, 8);

		this.mPhysicsWorld.setContactListener(this);

		/* Ball */
		this.mBallBody = PhysicsFactory.createCircleBody(this.mPhysicsWorld, 0,
				0, GameScene.BALL_SIDE/2,
				BodyType.DynamicBody, BALL_FIXTUREDEF);
		this.mBallBody.setBullet(true);

		/* Paddles */
		final Body paddleBodyBottom = PhysicsFactory.createBoxBody(this.mPhysicsWorld, GameScene.TABLE_WIDTH/2, 20, GameScene.PADDLE_WIDTH,
				GameScene.PADDLE_HEIGHT, BodyType.KinematicBody, PADDLE_FIXTUREDEF);
		paddleBodyBottom.setUserData(MainActivity.IS_SERVER);
		this.mPaddleBodies.put(MainActivity.IS_SERVER, paddleBodyBottom);

		final Body paddleBodyTop = PhysicsFactory.createBoxBody(this.mPhysicsWorld, GameScene.TABLE_WIDTH/2,
				GameScene.TABLE_HEIGHT-20, GameScene.PADDLE_WIDTH, GameScene.PADDLE_HEIGHT,
				BodyType.KinematicBody, PADDLE_FIXTUREDEF);
		paddleBodyTop.setUserData(MainActivity.IS_CLIENT);
		this.mPaddleBodies.put(MainActivity.IS_CLIENT, paddleBodyTop);

		this.initWalls();
	}
	
	public GameServer(int pServerPort, Engine pEngine, Options pOptions, boolean isSingle) {
//TODO change constant references to GameScene!!
		Log.i("pong", "Server constructor entered");
		this.mServerPort = pServerPort;
		this.mEngine = pEngine;
		this.mOptions = pOptions;
		this.mEngine.registerUpdateHandler(this);
		this.initMessagePool();

		this.mPaddleScores.put(MainActivity.IS_SERVER, new Score());
		this.mPaddleScores.put(MainActivity.IS_CLIENT, new Score());


		this.mPhysicsWorld = new FixedStepPhysicsWorld(MainActivity.FPS, 2, new Vector2(0, 0), false, 8, 8);

		this.mPhysicsWorld.setContactListener(this);

		/* Ball */
		this.mBallBody = PhysicsFactory.createCircleBody(this.mPhysicsWorld, 0,
				0, GameScene.BALL_SIDE/2,
				BodyType.DynamicBody, BALL_FIXTUREDEF);
		this.mBallBody.setBullet(true);

		/* Paddles */
		final Body paddleBodyBottom = PhysicsFactory.createBoxBody(this.mPhysicsWorld, GameScene.TABLE_WIDTH/2, 20, GameScene.PADDLE_WIDTH,
				GameScene.PADDLE_HEIGHT, BodyType.KinematicBody, PADDLE_FIXTUREDEF);
		paddleBodyBottom.setUserData(MainActivity.IS_SERVER);
		this.mPaddleBodies.put(MainActivity.IS_SERVER, paddleBodyBottom);

		
		if (!isSingle) {
			final Body paddleBodyTop = PhysicsFactory.createBoxBody(
					this.mPhysicsWorld, GameScene.TABLE_WIDTH / 2,
					GameScene.TABLE_HEIGHT - 20, GameScene.PADDLE_WIDTH,
					GameScene.PADDLE_HEIGHT, BodyType.KinematicBody,
					PADDLE_FIXTUREDEF);
			paddleBodyTop.setUserData(MainActivity.IS_CLIENT);
			this.mPaddleBodies.put(MainActivity.IS_CLIENT, paddleBodyTop);
		}
		else{
			final Body paddleBodyTop = PhysicsFactory.createBoxBody(
					this.mPhysicsWorld, GameScene.TABLE_WIDTH / 2,
					GameScene.TABLE_HEIGHT - 20, GameScene.TABLE_WIDTH,
					GameScene.PADDLE_HEIGHT, BodyType.KinematicBody,
					PADDLE_FIXTUREDEF);
			paddleBodyTop.setUserData(MainActivity.IS_CLIENT);
			this.mPaddleBodies.put(MainActivity.IS_CLIENT, paddleBodyTop);
		}
		this.initWalls();
	}
	
	public void initServer(){
		this.mEngine.runOnUpdateThread(new Runnable(){

			@Override
			public void run() {

				GameServer.this.mSocketServer = new SocketServer<SocketConnectionClientConnector>(
						GameServer.this.mServerPort, GameServer.this, GameServer.this) {
					protected SocketConnectionClientConnector newClientConnector(
							SocketConnection pSocketConnection) throws IOException {
						final SocketConnectionClientConnector clientConnector = new SocketConnectionClientConnector(pSocketConnection);
						clientConnector.registerClientMessage((short)GameClientMessages.CLIENT_MESSAGE_MOVE_PADDLE, ClientMsgMovePaddle.class, new IClientMessageHandler<SocketConnection>() {

							@Override
							public void onHandleMessage(final ClientConnector<SocketConnection> pClientConnector, final IClientMessage pClientMessage) throws IOException {
								final ClientMsgMovePaddle movePaddleClientMessage = (ClientMsgMovePaddle)pClientMessage;
								final Body paddleBody = GameServer.this.mPaddleBodies.get(movePaddleClientMessage.mPaddleID);
								final Vector2 paddlePosition = paddleBody.getTransform().getPosition();
								final float paddleX = MathUtils.bringToBounds(0, GameScene.TABLE_WIDTH, movePaddleClientMessage.mX);
								paddlePosition.set(paddleX / PhysicsConstants.PIXEL_TO_METER_RATIO_DEFAULT, paddlePosition.y);
								paddleBody.setTransform(paddlePosition, 0);
							}
						});
						
						clientConnector.registerClientMessage((short)GameClientMessages.CLIENT_MESSAGE_PAUSE_GAME, ClientMsgPauseGame.class, new IClientMessageHandler<SocketConnection>() {

							@Override
							public void onHandleMessage(final ClientConnector<SocketConnection> pClientConnector, final IClientMessage pClientMessage) throws IOException {
								final ClientMsgPauseGame pauseGameClientMessage = (ClientMsgPauseGame)pClientMessage;
								final boolean paused = pauseGameClientMessage.mPaused;
								mStartGame = !paused;
							}
						});
						
						clientConnector.registerClientMessage((short)GameClientMessages.CLIENT_MESSAGE_RESTART_GAME, ClientMsgRestartGame.class, new IClientMessageHandler<SocketConnection>() {

							@Override
							public void onHandleMessage(final ClientConnector<SocketConnection> pClientConnector, final IClientMessage pClientMessage) throws IOException {
								restartGame();
							}
						});
						
						clientConnector.registerClientMessage((short)GameClientMessages.CLIENT_MESSAGE_FINISH_GAME, ClientMsgFinishGame.class, new IClientMessageHandler<SocketConnection>() {

							@Override
							public void onHandleMessage(final ClientConnector<SocketConnection> pClientConnector, final IClientMessage pClientMessage) throws IOException {
								sendFinishMessage();
							}
						});
						return clientConnector;
					}
				};
				GameServer.this.mSocketServer.start();
				Log.i("pong", "Server started");

			}

		}, true);
	}
	
	private void initMessagePool() {
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_SCORE, ServerMsgUpdateScore.class);
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_BALL, ServerMsgUpdateBall.class);
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_PADDLE, ServerMsgUpdatePaddle.class);
		this.mMessagePool.registerMessage(GameServerMessages.SERVER_MESSAGE_FINISH_GAME, ServerMsgFinishGame.class);
	}
	
	private void initWalls() {
		WALL_FIXTUREDEF.isSensor = true;
		// goal lines
		final Body bottomGoalBody = PhysicsFactory.createLineBody(this.mPhysicsWorld,
				GameScene.TABLE_WIDTH, 1, 0, 1, WALL_FIXTUREDEF);
		bottomGoalBody.setUserData(this.mPaddleBodies.get(MainActivity.IS_SERVER));

		final Body topGoalBody = PhysicsFactory.createLineBody(this.mPhysicsWorld,
				0, GameScene.TABLE_HEIGHT, GameScene.TABLE_WIDTH, GameScene.TABLE_HEIGHT, WALL_FIXTUREDEF);
		topGoalBody.setUserData(this.mPaddleBodies.get(MainActivity.IS_CLIENT));
		this.mTopGoal = topGoalBody;


		WALL_FIXTUREDEF.isSensor = false;
		// side lines
		PhysicsFactory.createLineBody(this.mPhysicsWorld, 0, 0, 0, GameScene.TABLE_HEIGHT, WALL_FIXTUREDEF); //left line
		PhysicsFactory.createLineBody(this.mPhysicsWorld, GameScene.TABLE_WIDTH-1, GameScene.TABLE_HEIGHT,
				GameScene.TABLE_WIDTH-1, 0, WALL_FIXTUREDEF); //right line
	}

	@Override
	public void beginContact(Contact pContact) {
		final Fixture fixtureA = pContact.getFixtureA();
		final Body bodyA = fixtureA.getBody();
		final Object userDataA = bodyA.getUserData();

		final Fixture fixtureB = pContact.getFixtureB();
		final Body bodyB = fixtureB.getBody();
		final Object userDataB = bodyB.getUserData();

		final boolean isScoreSensorA = userDataA != null && userDataA instanceof Body;
		final boolean isScoreSensorB = userDataB != null && userDataB instanceof Body;
		Log.i("pong", "value bodyA = " + bodyA);
		Log.i("pong", "value bodyB = " + bodyB);

		if(isScoreSensorA || isScoreSensorB) {
			this.mResetBall = true;
						
			//Log.i("pong", "value bodyA = " + bodyA);
			//Log.i("pong", "value bodyB = " + bodyB);
			Log.i("pong", "value isScoreSensorA = " + isScoreSensorA);
			Log.i("pong", "value isScoreSensorB = " + isScoreSensorB);
			//final PaddleUserData paddleUserData = (isScoreSensorA) ? (PaddleUserData)(((Body)userDataA).getUserData()) : (PaddleUserData)(((Body)userDataB).getUserData());
			// Here we are!
			final int opponentID = mTopGoal.equals(bodyA) ? MainActivity.IS_SERVER : MainActivity.IS_CLIENT;
			final Score opponentPaddleScore = this.mPaddleScores.get(opponentID);
			opponentPaddleScore.increase();

			final ServerMsgUpdateScore updateScoreServerMessage = (ServerMsgUpdateScore)this.
					mMessagePool.obtainMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_SCORE);
			updateScoreServerMessage.set(opponentID, opponentPaddleScore.getScore());

			final SmartList<SocketConnectionClientConnector> clientConnectors = this.mSocketServer.mClientConnectors;
			for(int i = 0; i < clientConnectors.size(); i++) {
				try {
					final ClientConnector<SocketConnection> clientConnector = clientConnectors.get(i);
					clientConnector.sendServerMessage(updateScoreServerMessage);
				} catch (final Exception e) {
					Debug.e(e);
				}
			}
			this.mMessagePool.recycleMessage(updateScoreServerMessage);
		}
		
	}

	@Override
	public void endContact(Contact contact) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void preSolve(Contact contact, Manifold oldManifold) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void postSolve(Contact contact, ContactImpulse impulse) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void onUpdate(float pSecondsElapsed) {
		if(this.mStartGame){
			if(this.mResetBall) {
				this.mResetBall = false;
				final Vector2 vector2 = Vector2Pool.obtain(10, 10);
				this.mBallBody.setTransform(vector2, 0);

				//vector2.set(MathUtils.randomSign() * MathUtils.random(3, 4), MathUtils.randomSign() * MathUtils.random(3, 4));
				//vector2.set(MathUtils.random(3, 4), MathUtils.random(3, 4));
				vector2.set(MathUtils.random(3, 4), MathUtils.random(9, 10));
				
				int difficulty = mOptions.getDifficulty();
				
				switch(difficulty){
				case(Options.LOW):
					vector2.set(MathUtils.random(3, 4), MathUtils.random(3, 4));
					break;
				case(Options.MEDIUM):
					vector2.set(MathUtils.random(3, 4), MathUtils.random(9, 10));
					break;
				case(Options.HIGH):
					vector2.set(MathUtils.random(7, 9), MathUtils.random(13, 15));
					break;
				}
				
				this.mBallBody.setLinearVelocity(vector2);
				Vector2Pool.recycle(vector2);
			}
			this.mPhysicsWorld.onUpdate(pSecondsElapsed);
			//Log.i("pong", "onUpdate method entered in server");
			/* Prepare UpdateBallServerMessage. */
			final Vector2 ballPosition = this.mBallBody.getPosition();
			final float ballX = ballPosition.x * PhysicsConstants.PIXEL_TO_METER_RATIO_DEFAULT;// - GameScene.BALL_SIDE/2;
			//Log.i("pong", "ballX assigned");
			final float ballY = ballPosition.y * PhysicsConstants.PIXEL_TO_METER_RATIO_DEFAULT;// - GameScene.BALL_SIDE/2;
			//Log.i("pong", "ballY assigned");
			if(this.mMessagePool == null){
				Log.i("pong", "messagePool is null");
			}
			final ServerMsgUpdateBall updateBallServerMessage = (ServerMsgUpdateBall)this.mMessagePool.obtainMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_BALL);
			updateBallServerMessage.set(ballX, ballY);

			final ArrayList<ServerMsgUpdatePaddle> updatePaddleServerMessages = this.mUpdatePaddleServerMessages;

			/* Prepare UpdatePaddleServerMessages. */
			final SparseArray<Body> paddleBodies = this.mPaddleBodies;
			for(int j = 0; j < paddleBodies.size(); j++) {
				final int paddleID = paddleBodies.keyAt(j);
				final Body paddleBody = paddleBodies.get(paddleID);
				final Vector2 paddlePosition = paddleBody.getPosition();

				final float paddleX = paddlePosition.x * PhysicsConstants.PIXEL_TO_METER_RATIO_DEFAULT;// - PADDLE_WIDTH_HALF;
				final float paddleY = paddlePosition.y * PhysicsConstants.PIXEL_TO_METER_RATIO_DEFAULT;// - PADDLE_HEIGHT_HALF;

				final ServerMsgUpdatePaddle updatePaddleServerMessage = (ServerMsgUpdatePaddle)this.mMessagePool.obtainMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_PADDLE);
				updatePaddleServerMessage.set(paddleID, paddleX, paddleY);

				updatePaddleServerMessages.add(updatePaddleServerMessage);
			}

			try {
				/* Update Ball. */
				this.mSocketServer.sendBroadcastServerMessage(updateBallServerMessage);

				/* Update Paddles. */
				for(int j = 0; j < updatePaddleServerMessages.size(); j++) {
					this.mSocketServer.sendBroadcastServerMessage(updatePaddleServerMessages.get(j));
				}
				//this.mSocketServer.sendBroadcastServerMessage(updateBallServerMessage);
			} catch (final Exception e) {
				Debug.e(e);
			}

			/* Recycle messages. */
			this.mMessagePool.recycleMessage(updateBallServerMessage);
			this.mMessagePool.recycleMessages(updatePaddleServerMessages);
			updatePaddleServerMessages.clear();
		}
	}

	@Override
	public void reset() {
		// TODO Auto-generated method stub
		
	}
	
	public void terminate(){
		mSocketServer.terminate();
	}
	
	public void setStartGame(boolean pValue){
		mStartGame = pValue;
	}
	
	public void restartGame(){
		mStartGame = true;
		mResetBall = true;
		for(int i = 0; i < 2; i++) {
			mPaddleScores.get(i).reset();
		}
		final ServerMsgUpdateScore updateScoreServerMessage = (ServerMsgUpdateScore)this.
				mMessagePool.obtainMessage(GameServerMessages.SERVER_MESSAGE_UPDATE_SCORE);
		updateScoreServerMessage.set(GameActivity.IS_SERVER, 0);

		final SmartList<SocketConnectionClientConnector> clientConnectors = this.mSocketServer.mClientConnectors;
		
		for(int i = 0; i < clientConnectors.size(); i++) {
			try {
				final ClientConnector<SocketConnection> clientConnector = clientConnectors.get(i);
				clientConnector.sendServerMessage(updateScoreServerMessage);
			} catch (final Exception e) {
				Debug.e(e);
			}
		}
		updateScoreServerMessage.set(GameActivity.IS_CLIENT, 0);
		for(int i = 0; i < clientConnectors.size(); i++) {
			try {
				final ClientConnector<SocketConnection> clientConnector = clientConnectors.get(i);
				clientConnector.sendServerMessage(updateScoreServerMessage);
			} catch (final Exception e) {
				Debug.e(e);
			}
		}
	}
	
	public void sendFinishMessage(){
		final ServerMsgFinishGame finishGameServerMessage = (ServerMsgFinishGame)this.
				mMessagePool.obtainMessage(GameServerMessages.SERVER_MESSAGE_FINISH_GAME);
		this.mSocketServer.sendBroadcastServerMessage(finishGameServerMessage);
	}
	

	@Override
	public void onStarted(ClientConnector<SocketConnection> pClientConnector) {
		Log.i("pong", "SERVER: connection accepted");
		
	}

	@Override
	public void onTerminated(ClientConnector<SocketConnection> pClientConnector) {
		//pClientConnector.terminate();
		
	}

	@Override
	public void onStarted(
			SocketServer<SocketConnectionClientConnector> pSocketServer) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void onTerminated(
			SocketServer<SocketConnectionClientConnector> pSocketServer) {
		//pSocketServer.terminate();
		
	}

	@Override
	public void onException(
			SocketServer<SocketConnectionClientConnector> pSocketServer,
			Throwable pThrowable) {
		// TODO Auto-generated method stub
		
	}
}
